//
// Copyright  2003-2010, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// BotVoiceCodecDLLWrapper.hpp
//
// Class: VoiceCodecDLLWrapper
//
// Version: $ID:$
//

#if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000
#	pragma once
#endif	// if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000

#ifndef BOT_VOICE_CODEC_DLL_WRAPPER_INCLUDED
#define BOT_VOICE_CODEC_DLL_WRAPPER_INCLUDED

inline const bool IsDirectoryExists (const char *const folderName)
{
	const int directoryAttributes (GetFileAttributes (folderName));

	return directoryAttributes != INVALID_FILE_ATTRIBUTES && (directoryAttributes & FILE_ATTRIBUTE_DIRECTORY);
}
inline const bool CreateDirectory (const char *const directory)
{
	// This function creates a directory, by given 'directory' string.

	return mkdir (directory) == 0/* || errno == EEXIST*/;
}
static const bool MakeLocalFileCopy (const char *const name)
{
	// test if the game.DLL file is NOT available outside of the steam cache
	if (access (name, 0) == 0)
		return true;

	// try to extract the game.DLL out of the steam cache
	(*g_engfuncs.pfnServerPrint) (va ("Warning: Trying to extract '%s' out of the steam cache....\n", name));	// print it

	// is directory for new game.DLL not exists?
	if (!IsDirectoryExists (name))
	{
		(*g_engfuncs.pfnServerPrint) (va ("Directory for new game.DLL '%s' not created, trying to create it....\n", name));	// print it

		// is directory for new game.DLL not created?
		if (!CreateDirectory (name))
		{
			(*g_engfuncs.pfnServerPrint) (va ("Error: Couldn't create a new game.DLL '%s' directory.\n", name));

			return false;
		}
	}

	unsigned int steamFileSize;
	unsigned char *const steamFileBuffer (LOAD_FILE_FOR_ME (name, &steamFileSize));	// open game.DLL from steam cache

	// is game.DLL found?
	assert (steamFileBuffer != NULL);

	// create and open new game.DLL in binary write mode
	FILE *const localFile (fopen (name, "wb"));

	// is file created?
	if (localFile == NULL)
	{
		// Free used file...
		FREE_FILE (steamFileBuffer);

		(*g_engfuncs.pfnServerPrint) (va ("Error: Couldn't create a new '%s' file.\n", name));

		return false;
	}

	// if in cache, then extract it
	if (fwrite (steamFileBuffer, steamFileSize, 1u, localFile) != 1u)
	{
		// Free used file...
		FREE_FILE (steamFileBuffer);

		(*g_engfuncs.pfnServerPrint) (va ("Error: Couldn't write buffer into a new '%s' file.\n", name));

		return false;
	}

	// and free used file...
	FREE_FILE (steamFileBuffer);

	return true;	// Done.
}

class VoiceCodecDLLWrapper
{
	//
	// Group: Private members.
	//
	private:
		enum { MaximumVoiceCodecNameLength = 64u };

		#include <VoiceCodecLibrary.hpp>

		CodecLibrary  m_voiceCodecDLL;
		char          m_codecName[MaximumVoiceCodecNameLength];
		unsigned char m_quality;

	//
	// Group: (Con/De)structors.
	//
	public:
		inline VoiceCodecDLLWrapper (void) :
			m_quality (0u)
		{
			m_codecName[0u] = '\0';

			if (sv_voiceenable->value != 0.0f)
				Initialize (sv_voicecodec->string);
		}

	//
	// Group: Functions.
	//
	public:
		inline const char *const    GetCodecName     (void) const { return m_codecName; }
		inline const CodecLibrary  &GetVoiceCodecDLL (void) const { return m_voiceCodecDLL; }
		inline const unsigned char  GetQuality       (void) const { return m_quality; }

		inline const bool Initialized (void) const { return m_voiceCodecDLL.IsValid () && m_voiceCodecDLL.GetCreateCodecFunction () != NULL; }
		inline const bool Initialize (const char *const codecName)
		{
			// Reliability checks.
			assert (codecName != NULL && codecName[0u] != '\0');
			assert (!Initialized ());

			// Get the codec name with extention.
			const char codecDLLName[MaximumVoiceCodecNameLength];
			_snprintf (codecDLLName, sizeof (codecDLLName), "%s.DLL", codecName);

			MakeLocalFileCopy ("MSS32.DLL");	// voice_miles.DLL will use this...
			MakeLocalFileCopy (codecDLLName);

			if (!m_voiceCodecDLL.Load (codecDLLName))
			{
				(*g_engfuncs.pfnServerPrint) (va ("Error: Unable to load voice codec DLL '%s'. Bot voice disabled.\n", codecName));

				return false;
			}

			if (!m_voiceCodecDLL.InitializeCreateCodecFunction ())
			{
				(*g_engfuncs.pfnServerPrint) (va ("Error: Unable to get a factory for voice codec '%s'. Bot voice disabled.\n", codecName));

				return false;
			}

			strncpy (m_codecName, codecName, sizeof (m_codecName));
			m_quality = static_cast <unsigned char> (sv_voicequality->value);	// Update quality value....

			(*g_engfuncs.pfnServerPrint) (va ("Voice codec '%s' successfully loaded. (Base address = 0x%X)\n", codecDLLName, m_voiceCodecDLL.GetPointer ()));

			return true;
		}
		inline void Deinitialize (void)
		{
			// Reliability check.
			assert (Initialized ());

			ReleaseCodecForBots ();

			m_voiceCodecDLL.Free ();
		}

		inline const bool ChangeCodec (const char *const newCodecName)
		{
			if (Initialized ())
				Deinitialize ();

			if (!Initialize (newCodecName))
				return false;

			OnChangeCodecForBots (GetQuality ());

			return true;
		}
		inline void ChangeQuality (const unsigned char newQuality)
		{
			if (!Initialized ())
				return;

			m_quality = newQuality;	// Update quality value....

			ReleaseCodecForBots ();

			OnChangeCodecForBots (newQuality);
		}

		inline IVoiceCodec *const InitializeCodec (const unsigned char quality) const
		{
			// Reliability check.
			assert (Initialized ());

			IVoiceCodec *const newCodec (static_cast <IVoiceCodec *> ((*m_voiceCodecDLL.GetCreateCodecFunction ()) (GetCodecName (), NULL)));

			if (newCodec == NULL)
			{
				(*g_engfuncs.pfnServerPrint) (va ("Error: Unable to create interface for voice codec '%s'. Bot voice disabled.\n", GetCodecName ()));

				return NULL;
			}

			if (!newCodec->Init (quality))
			{
				(*g_engfuncs.pfnServerPrint) (va ("Error: Unable to initialize voice codec '%s'. Bot voice disabled.\n", GetCodecName ()));

				// The commented code is used in the engine to resolve conflicts between
				// the codec and the class 'CDAudio' at the moment of unloading of the codec module
				// and also when calling the function 'IVoiceCodec::Release',
				// because both classes ('IVoiceCodec' and 'CDAudio') use the same library of "MSS" (Miles Sound System).
				// It is actual only for codec "voice_miles"!!!
				// However, I commented that code because I wanted to avoid wild hooks....
//				if (IS_DEDICATED_SERVER ())
					newCodec->Release ();
/*				else
				{
					CDAudio *const cdaudio (...);

					cdaudio->UNKNOWN2 ();

						newCodec->Release ();

					cdaudio->UNKNOWN1 ();
				}
*/
				return NULL;
			}

			(*g_engfuncs.pfnServerPrint) (va ("Voice codec '%s' successfully initialized. Bot voice enabled.\n", GetCodecName ()));

			return newCodec;
		}
		inline IVoiceCodec *const TryInitializeCodec (const unsigned char quality) const
		{
			if (sv_voiceenable->value == 0.0f)
				return NULL;

			return InitializeCodec (quality);
		}

		inline void SV_CheckVoiceChanges (void)
		{
			if (stricmp (m_codecName, sv_voicecodec->string) != 0)
			{
				ChangeCodec (voice_codec);

				return;
			}

			if (GetQuality () != static_cast <unsigned char> (sv_voicequality->value))
				ChangeQuality (static_cast <unsigned char> (sv_voicequality->value));
		}
};

#endif	// ifndef BOT_VOICE_CODEC_DLL_WRAPPER_INCLUDED