//
// Copyright  2003-2010, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// VoiceCodecLibrary.hpp
//
// Class: VoiceCodecLibrary
//
// Version: $ID:$
//

#if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000
#	pragma once
#endif	// if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000

#ifndef VOICE_CODEC_LIBRARY_INCLUDED
#define VOICE_CODEC_LIBRARY_INCLUDED

class VoiceCodecLibrary
{
	//
	// Group: Private members.
	//
	private:
		union Module_t
		{
			HMODULE           handle;
			void             *pointer;
			char             *characterBuffer;
			unsigned char    *byteBuffer;
			IMAGE_DOS_HEADER *DOSHeader;

			inline Module_t (HMODULE inputHandle = NULL) : handle (inputHandle) { /* VOID */ }

			// casting operators
			inline operator       HMODULE (void)       { return handle; }	// Library will now automatically convert to 'HMODULE *' when needed
			inline operator const HMODULE (void) const { return handle; }	// Library will now automatically convert to 'const HMODULE *const' when needed
		} m_module;	// Handle to the loaded library. (Specific handle that points to dynamic library (This comes from LoadLibrary() on Windows))

		CreateInterfaceFn m_createCodecFunction;	// Create factory.

	//
	// Group: (Con/De)structors.
	//
	public:
		inline VoiceCodecLibrary (HMODULE pointer = NULL) : m_module (pointer), m_createCodecFunction (NULL) { /* VOID */ }
		inline VoiceCodecLibrary (const VoiceCodecLibrary &right);

		inline ~VoiceCodecLibrary (void)
		{
			if (IsValid ())
				FreeLibrary (m_module);
		}

	//
	// Group: Operators.
	//
	public:
		// casting operators
		inline operator       HMODULE (void)       { return m_module; }	// Library will now automatically convert to 'HMODULE' when needed
		inline operator const HMODULE (void) const { return m_module; }	// Library will now automatically convert to 'const HMODULE' when needed
		inline operator const bool    (void) const { return IsValid (); }

		inline VoiceCodecLibrary &operator = (HMODULE pointer)
		{
			m_module = pointer;

			// Reliability check.
			InternalAssert (IsValid ());

			return *this;
		}
		inline VoiceCodecLibrary &operator = (const VoiceCodecLibrary &right);

	//
	// Group: Functions.
	//
	public:
		inline const bool IsValid (void) const { return m_module.handle != NULL; }

		inline       HMODULE &GetPointer  (void)       { return m_module.handle; }	// Get the pointer of the library.
		inline const HMODULE &GetPointer  (void) const { return m_module.handle; }	// Get the pointer of the library.

		//
		// Function: GetFunctionAddress
		//
		// Description: Gets the address of the given symbol from the loaded dynamic library.
		//
		// Parameters:
		//	functionName - Name of the symbol to search for.
		//
		// Returns: If the function succeeds, the returned value is a pointer to the symbol, NULL otherwise.
		//
		template <typename elementType> inline const elementType GetFunctionAddress (const char *const functionName) const
		{
			return reinterpret_cast <elementType> (GetProcAddress (m_module, functionName));
		}

		inline const CreateInterfaceFn GetCreateCodecFunction (void) const { return m_createCodecFunction; }

		inline const bool Load (const char *const pathName)
		{
			// Reliability check.
			assert (!IsValid ());

			m_module = LoadLibrary (pathName);

			return IsValid ();
		}
		inline const bool InitializeCreateCodecFunction (void)
		{
			// Reliability check.
			assert (IsValid ());

			m_createCodecFunction = GetFunctionAddress <CreateInterfaceFn> (CREATEINTERFACE_PROCNAME);

			return m_createCodecFunction != NULL;
		}
		inline const bool Free (void)
		{
			// Reliability check.
			if (!IsValid ())
				return true;

			bool result;

			// The commented code is used in the engine to resolve conflicts between
			// the codec and the class 'CDAudio' at the moment of unloading of the codec module
			// and also when calling the function 'IVoiceCodec::Release',
			// because both classes ('IVoiceCodec' and 'CDAudio') use the same library of "MSS" (Miles Sound System).
			// It is actual only for codec "voice_miles"!!!
			// However, I commented that code because I wanted to avoid wild hooks here....
//			if (IS_DEDICATED_SERVER ())
				result = FreeLibrary (m_module) == TRUE;
/*			else
			{
				CDAudio *const cdaudio (...);

				cdaudio->UNKNOWN2 ();

					result = FreeLibrary (m_module) == TRUE;

				cdaudio->UNKNOWN1 ();
			}
*/
			m_module.handle = NULL;
			m_createCodecFunction = NULL;

			return result;
		}
};

#endif	// ifndef VOICE_CODEC_LIBRARY_INCLUDED