//
// Copyright  2003-2010, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// WaveFileData.hpp
//
// Class: WaveSoundData
//
// Version: $ID:$
//

#if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000
#	pragma once
#endif	// if defined COMPILER_VISUALC && COMPILER_VISUALC > 1000

#ifndef WAVE_FILE_DATA_INCLUDED
#define WAVE_FILE_DATA_INCLUDED

// WAVE file related structure definitions
struct WaveFileHeader_t
{
	struct WaveFormat_t
	{
/*! Off=0(0)/20(5) */	unsigned short tag;					// format type of the sound data (1 - simple, 0x101 - IBM mu-law, 0x102 - IBM a-law, 0x103 - ADPCM)
/*! Off=2(0.5)/22(5.5) */	unsigned short channels;			// number of channels (0x01 - mono, 0x02 - stereo)
/*! Off=4(1)/24(6) */	unsigned int   sampleRate;			// sample rate in hertz (num samples per second (e.g. 11025 - 11 khz))
/*! Off=8(2)/28(7) */	unsigned int   bytesPerSecond;		// bytes of data read per second (bps) when playing the file
/*! Off=12(3)/32(8) */	unsigned short bytesPerSample;		// bytes per sample (including all channels)
/*! Off=14(3.5)/34(8.5) */	unsigned short bitsPerSample;	// bits per sample, bytesPerSample * 8
	};	// sizeof (WaveFormat_t) == 16 (4)

	// RIFF chunk
/*! Off=0(0) */	char           riffChunkID[GET_STATIC_STRING_LENGTH ("RIFF")];		// RIFF chunk ID ("RIFF")
/*! Off=4(1) */	unsigned int   packageSize;											// total length of package (all following chunks) in bytes

	// WAVE chunk
/*! Off=8(2) */	char           waveChunkID[GET_STATIC_STRING_LENGTH ("WAVE")];		// WAVE chunk ID ("WAVE")
/*! Off=12(3) */	char           formatChunkID[GET_STATIC_STRING_LENGTH ("fmt ")];	// FORMAT chunk ID ("fmt ")
/*! Off=16(4) */	unsigned int   formatChunkLength;									// length of FORMAT chunk in bytes

	// WAVE ID
/*! Off=20(5) */	WaveFormat_t   format;

	// DATA chunk
/*! Off=36(9) */	char           dataChunkID[GET_STATIC_STRING_LENGTH ("data")];		// DATA chunk ID ("data")
/*! Off=40(10) */	unsigned int   dataChunkLength;										// length of DATA chunk in bytes
};	// sizeof (WaveFileHeader_t) == 44 (11)

class WaveSoundData
{
	private:
		// current sound state
		WaveFileHeader_t  m_header;
		size_t            m_size;	// sound unpacked size (for bounds checking)
		unsigned char    *m_data;	// sound pointer (see 'm_header.format.tag' for details)

	public:
		inline WaveSoundData (void) : m_data (NULL), m_size (0u) { memset (&m_header, 0u, sizeof (m_header)); }
		inline WaveSoundData (FILE *const file)
		{
			// Reliability check.
			assert (file != NULL);

			if (fread (&m_header, sizeof (m_header), 1u, file) != 1)
				return;

			m_size = m_header.dataChunkLength;

			// Reliability check.
			if (m_size == 0u)
				return;

			m_data = new unsigned char[m_size];

			// Load the data
			if (fread (m_data, m_size, 1u, file) != 1)
			{
				Delete ();

				m_data = NULL;

				return;
			}

			Initialize ();
		}
		inline WaveSoundData (const unsigned char *const buffer, const unsigned short size) :
			m_header (*reinterpret_cast <const WaveFileHeader_t *const> (buffer)),
			m_size (m_header.dataChunkLength),
			m_data (new unsigned char[m_size])
		{
			// Reliability checks.
			assert (buffer != NULL);
			assert (size >= sizeof (m_header) + m_size);

			memcpy (m_data, buffer, m_size);	// Load the data

			Initialize ();
		}
		inline WaveSoundData (const WaveSoundData &soundData) :
			m_header (soundData.m_header),
			m_size (m_header.dataChunkLength),
			m_data (new unsigned char [m_size])
		{
			// Load the data
			memcpy (m_data, soundData.m_data, m_size);
		}

		inline ~WaveSoundData (void) { Delete (); }

	protected:
		inline const bool Merge8 (const WaveSoundData &otherSoundData)
		{
			return m_header.format.channels == 1u ? Merge8Mono (otherSoundData) : Merge8Stereo (otherSoundData);
		}
		inline const bool Merge8Mono (const WaveSoundData &otherSoundData)
		{
			if (otherSoundData.GetHeader ().format.bitsPerSample != GetHeader ().format.bitsPerSample || otherSoundData.GetHeader ().format.sampleRate != GetHeader ().format.sampleRate)
				return false;

			unsigned char *const commonData (new unsigned char[m_size + otherSoundData.m_size]);

			if (commonData == NULL)
				return false;

			memcpy (commonData, m_data, m_size);

			delete [] m_data;

			memcpy (commonData + m_size, otherSoundData.m_data, otherSoundData.m_size);

			m_data = commonData;

			m_size += otherSoundData.m_size;
			m_header.dataChunkLength += m_header.dataChunkLength;

			return true;
		}
		inline const bool Merge8Stereo (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!
		inline const bool Merge16 (const WaveSoundData &otherSoundData)
		{
			return m_header.format.channels == 1u ? Merge16Mono (otherSoundData) : Merge16Stereo (otherSoundData);
		}
		inline const bool Merge16Mono (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!
		inline const bool Merge16Stereo (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!

		inline const bool Mix8 (const WaveSoundData &otherSoundData)
		{
			return m_header.format.channels == 1u ? Mix8Mono (otherSoundData) : Mix8Stereo (otherSoundData);
		}
		inline const bool Mix8Mono (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!
		inline const bool Mix8Stereo (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!
		inline const bool Mix16 (const WaveSoundData &otherSoundData)
		{
			return m_header.format.channels == 1u ? Mix16Mono (otherSoundData) : Mix16Stereo (otherSoundData);
		}
		inline const bool Mix16Mono (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!
		inline const bool Mix16Stereo (const WaveSoundData &otherSoundData) { return false; }	/// @todo !!!

		inline void SetVolume8 (const float newVolume)
		{
			if (m_header.format.channels == 1u)
				SetVolume8Mono (newVolume);
			else
				SetVolume8Stereo (newVolume);
		}
		inline void SetVolume8Mono (const float newVolume)
		{
			const unsigned short volume (static_cast <unsigned short> (newVolume * 256u));

			for (size_t index (0u); index < m_size; ++index)
				m_data[index] = static_cast <char> (Math::GetClampedValueInRange <int> ((m_data[index] * volume) >> 8u/* static_cast <int> (m_data[index] * newVolume) */, -256, 255));
		}
		inline void SetVolume8Stereo (const float newVolume) { /*! @todo !!! */ }
		inline void SetVolume16 (const float newVolume)
		{
			if (m_header.format.channels == 1u)
				SetVolume16Mono (newVolume);
			else
				SetVolume16Stereo (newVolume);
		}
		inline void SetVolume16Mono (const float newVolume)
		{
			short *tempShortData (reinterpret_cast <short *> (m_data));
			const short *const tempShortDataEnd (tempShortData + m_size / sizeof (*tempShortData));
			const unsigned short volume (static_cast <unsigned short> (newVolume * 256u));

			for (/* Empty */; tempShortData < tempShortDataEnd; ++tempShortData)
				*tempShortData = static_cast <short> (Math::GetClampedValueInRange <int> ((*tempShortData * volume) >> 8u/* static_cast <int> (*tempShortData * newVolume) */, -32768, 32767));
		}
		inline void SetVolume16Stereo (const float newVolume) { /*! @todo !!! */ }

		inline void ConvertToSigned (void)
		{
			// Do any sample conversion for 8 bit PCM, convert to signed because the mixing routine assumes this.

			if (m_header.format.bitsPerSample == 8u)
				ConvertToSigned8 ();	// now convert 8-bit sounds to signed
			else
				ConvertToSigned16 ();	// now convert 16-bit sounds to signed
		}
		inline void ConvertToSigned8 (void)
		{
			// Convert unsigned data to signed.

			for (size_t index (0u); index < m_size; ++index)
				reinterpret_cast <char *> (m_data)[index] = static_cast <char> (static_cast <int> (m_data[index] - 128));
		}
		inline void ConvertToSigned16 (void)
		{
			// Convert unsigned data to signed.

			for (size_t index (0u); index < m_size; ++index)
			{
				reinterpret_cast <char *> (m_data)[index * 2u + 0u] = static_cast <char> (static_cast <int> (m_data[index * 2u + 0u] - 128));
				reinterpret_cast <char *> (m_data)[index * 2u + 1u] = static_cast <char> (static_cast <int> (m_data[index * 2u + 1u] - 128));
			}
/*			const size_t nSamples (m_size / sizeof (short));
			short *const pDest (reinterpret_cast <short *> (m_data));
			for (size_t i (0u); i < nSamples; ++i)
			{
				const int val = *((unsigned short*)&pDest[i]) - (1 << 15);
				pDest[i] = (short)val;
			}*/
		}
		inline void ConvertToUnsigned (void)
		{
			if (m_header.format.bitsPerSample == 8u)
				ConvertToUnsigned8 ();	// now convert 8-bit sounds to unsigned
			else
				ConvertToUnsigned16 ();	// now convert 16-bit sounds to unsigned
		}
		inline void ConvertToUnsigned8 (void)
		{
			// Convert signed data to unsigned.

			for (size_t index (0u); index < m_size; ++index)
				reinterpret_cast <unsigned char *> (m_data)[index] = static_cast <unsigned char> (static_cast <int> (m_data[index] + 128));
		}
		inline void ConvertToUnsigned16 (void)
		{
			// Convert signed data to unsigned.

			for (size_t index (0u); index < m_size; ++index)
			{
				reinterpret_cast <unsigned char *> (m_data)[index * 2u + 0u] = static_cast <unsigned char> (static_cast <int> (m_data[index * 2u + 0u] + 128));
				reinterpret_cast <unsigned char *> (m_data)[index * 2u + 1u] = static_cast <unsigned char> (static_cast <int> (m_data[index * 2u + 1u] + 128));
			}
/*			const size_t nSamples (m_size / sizeof (short));
			short *const pDest (reinterpret_cast <short *> (m_data));
			for (size_t i (0u); i < nSamples; ++i)
			{
				const int val = *((short*)&pDest[i]) + (1 << 15);
				pDest[i] = (unsigned short)val;
			}*/
		}

	public:
		inline const bool Merge (const WaveSoundData &otherSoundData)
		{
			return m_header.format.bitsPerSample == 8u ? Merge8 (otherSoundData) : Merge16 (otherSoundData);
		}
		inline const bool Merge (FILE *const otherSoundFile)
		{
			// Reliability check.
			assert (otherSoundFile != NULL);

			WaveSoundData otherSoundData (otherSoundFile);

			if (!otherSoundData.IsValid ())
				return false;

			return Merge (otherSoundData);
		}
		inline const bool Mix (const WaveSoundData &otherSoundData)
		{
			return m_header.format.bitsPerSample == 8u ? Mix8 (otherSoundData) : Mix16 (otherSoundData);
		}
		inline const bool Mix (FILE *const otherSoundFile)
		{
			// Reliability check.
			assert (otherSoundFile != NULL);

			WaveSoundData otherSoundData (otherSoundFile);

			if (!otherSoundData.IsValid ())
				return false;

			return Mix (otherSoundData);
		}
		inline void SetVolume (const float newVolume)
		{
			// Fast case - nothing changes.
			if (newVolume == 1.0f)
				return;

			if (m_header.format.bitsPerSample == 8u)
				SetVolume8 (newVolume);
			else
				SetVolume16 (newVolume);
		}

		inline const bool IsValid (void) const { return m_data != NULL; }
		inline void Delete (void) { delete [] m_data; }
		inline       WaveFileHeader_t &GetHeader (void)       { return m_header; }
		inline const WaveFileHeader_t &GetHeader (void) const { return m_header; }
		inline       unsigned char *const GetSamples (void)       { return m_data; }
		inline const unsigned char *const GetSamples (void) const { return m_data; }
		inline const size_t GetSamplesSize (void) const { return m_size; }
		inline const float GetDuration (void) const
		{
			// now return the duration
			return static_cast <float> (GetHeader ().dataChunkLength) / ((GetHeader ().format.bitsPerSample * GetHeader ().format.channels) / 8u) / GetHeader ().format.sampleRate;	// SOURCE ENGINE
//			return static_cast <float> (GetHeader ().dataChunkLength) / static_cast <float> (GetHeader ().format.bytesPerSecond);	// ORIGINAL
/*
			const float secondLength (static_cast <float> (GetHeader ().dataChunkLength) / static_cast <float> (GetHeader ().format.bytesPerSecond));

			if (!Math::IsZero (secondLength))
				return secondLength;

			char ch[32u];

			_snprintf (ch, sizeof (ch), "0.%u", GetHeader ().dataChunkLength);

			// else return milli second length.
			return static_cast <float> (atof (ch));*/
		}

		inline void Initialize (void)
		{
			if (m_header.format.bitsPerSample == 8u)
				ConvertToSigned8 ();	// now convert 8-bit sounds to signed
		}

		inline const bool ResampleInternal (const unsigned short outSampleRate = 8000u, const unsigned char outBitsPerSample = 16u)
		{
			// We need convert sound to signed even if nothing to resample.

			const unsigned char outWidth (static_cast <unsigned char> (outBitsPerSample / 8u));	// resolution - bum bits divided by 8 (8 bit is 1, 16 bit is 2)
			const unsigned char width (static_cast <unsigned char> (m_header.format.bitsPerSample / 8u));	// resolution - bum bits divided by 8 (8 bit is 1, 16 bit is 2)
			const unsigned int samples (m_header.dataChunkLength / width / m_header.format.channels);	// total samplecount in sound
			int srcsample;
			int i, sample, sample2, samplefrac, fracstep;
			const float stepscale (static_cast <float> (m_header.format.sampleRate) / outSampleRate);	// this is usually 0.5, 1, or 2
			const int outcount (static_cast <int> (samples / stepscale));

			m_header.dataChunkLength = m_size = outcount * outWidth * m_header.format.channels;

			unsigned char *const tempbuffer (new unsigned char[m_size]);
/*
			samples = outcount;
			if (sound.loopstart != -1)
				sound.loopstart = static_cast <int> (sound.loopstart / stepscale);
*/
			// resample / decimate to the current source rate
			if (stepscale == 1.0f && width == 1u && outWidth == 1u)
			{
				// fast special case
				ConvertToSigned8 ();
			}
			else
			{
				// general case
				samplefrac = 0;
				fracstep = static_cast <int> (stepscale * 256);

				if (m_header.format.channels == 2u)
				{
					for (i = 0; i < outcount; ++i)
					{
						srcsample = samplefrac >> 8;
						samplefrac += fracstep;

						if (width == 2u)
						{
							sample = reinterpret_cast <short *> (m_data)[srcsample * 2u + 0u];
							sample2 = reinterpret_cast <short *> (m_data)[srcsample * 2u + 1u];
						}
						else
						{
							sample = static_cast <int> (static_cast <char> (m_data[srcsample * 2u + 0u])) << 8;
							sample2 = static_cast <int> (static_cast <char> (m_data[srcsample * 2u + 1u])) << 8;
						}

						if (outWidth == 2u)
						{
							reinterpret_cast <short *> (tempbuffer)[i * 2u + 0u] = static_cast <short> (sample);
							reinterpret_cast <short *> (tempbuffer)[i * 2u + 1u] = static_cast <short> (sample2);
						}
						else
						{
							reinterpret_cast <char *> (tempbuffer)[i * 2u + 0u] = static_cast <char> (sample >> 8);
							reinterpret_cast <char *> (tempbuffer)[i * 2u + 1u] = static_cast <char> (sample2 >> 8);
						}
					}
				}
				else for (i = 0; i < outcount; ++i)
				{
					srcsample = samplefrac >> 8;
					samplefrac += fracstep;

					if (width == 2u)
						sample = reinterpret_cast <short *> (m_data)[srcsample];
					else
						sample = static_cast <int> (static_cast <char> (m_data[srcsample])) << 8;

					if (outWidth == 2u)
						reinterpret_cast <short *> (tempbuffer)[i] = static_cast <short> (sample);
					else
						reinterpret_cast <char *> (tempbuffer)[i] = static_cast <char> (sample >> 8);
				}
			}

			m_header.format.sampleRate = outSampleRate;
			m_header.format.bitsPerSample = outBitsPerSample;
			m_header.format.bytesPerSample = m_header.format.channels * m_header.format.bitsPerSample / 8u;
			m_header.format.bytesPerSecond = m_header.format.sampleRate * m_header.format.bytesPerSample;

			delete [] m_data;
			m_data = tempbuffer;

			return true;
		}

	public:
		inline WaveSoundData &operator = (FILE *const file)
		{
			// Reliability check.
			assert (file != NULL);

			Delete ();

			fread (&m_header, sizeof (m_header), 1u, file);

			m_size = m_header.dataChunkLength;
			m_data = new unsigned char[m_size];

			// Load the data
			fread (m_data, m_size, 1u, file);

			Initialize ();

			return *this;
		}
		inline WaveSoundData &operator = (const WaveSoundData &soundData)
		{
			Delete ();

			m_header = soundData.m_header;
			m_size = m_header.dataChunkLength;
			m_data = new unsigned char [m_size];

			// Load the data
			memcpy (m_data, soundData.m_data, m_size);

			return *this;
		}
};

#endif	// ifndef WAVE_FILE_DATA_INCLUDED